import java.util.*;
import java.awt.Font;
import java.net.URL;
import java.io.File;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Point;
//import java.awt.Rectangle;
import java.awt.event.*;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.Highlighter;

import net.sourceforge.nite.search.*;
import net.sourceforge.nite.meta.*;
import net.sourceforge.nite.meta.impl.*;
import net.sourceforge.nite.nom.nomwrite.*;
import net.sourceforge.nite.nom.nomwrite.impl.*;
import net.sourceforge.nite.gui.util.*;
import net.sourceforge.nite.query.*;
import net.sourceforge.nite.gui.textviewer.*;
import net.sourceforge.nite.gui.mediaviewer.*;
import net.sourceforge.nite.nxt.*;
import net.sourceforge.nite.nstyle.handler.*;
import net.sourceforge.nite.time.*;
import net.sourceforge.nite.gui.actions.ChangeAttributeValueOptionPane;

import net.sourceforge.nite.nom.NOMException;
import javax.swing.plaf.metal.*;

/**
 * Switchboard general display.
 *
 * Markables are denoted by parentheses.  The colour of the
 * bracketing is determined by the nesting level, to make it
 * easy to match up brackets visually.  This is badly hacked from
 * one of the original coders - do not use it as model NXT programming.
 * 
 * Oct 2003, Jean Carletta, modifying so that there is an option to
 * show syntax trees
 * 
 * May 2008, updating to show orthography for public release, show POS
 * tags, and use a modern search window - otherwise unchanged.
 * 
 * There is a bug that the first anaphor on the link window won't highlight
 * the corresponding part of the main display on its own.
 * 
 * @author Jonathan Kilgour, Jean Carletta April 2003
 **/

public class SwitchboardDisplay
	implements ActionListener, MouseListener, WindowListener, TreeSelectionListener {

	/* SOME CONSTANTS that make it easier to adapt this code */
	//the exact name of the element being marked up
	private final String MARKABLE = "markable";

	//the exact names of the attributes being marked up
	// (not all of these get used in this program; some are
	// for the Edinburgh coder)
	private final String ANIMACY = "animacy";
	private final String ANTHRO = "anthro";
	private final String ANIMCONF = "animconf";
	private final String STATUS = "status";
	private final String STATUSTYPE = "statustype";
	// what to put on the button for adding a note
	private final String NOTEBUTTON = "Add note";
	// the name of the attribute for storing notes
	private final String NOTE = "edin-note";

	// the attribute values for these attributes we wish to
	// treat as privileged, meaning "uncoded".  Uncoded means
	// that the coder hasn't considered the value yet - the attribute
	// value is actually missing if they haven't considered it, because
	// that's the way the initial data was imported, but this explicit
	// value allows them to say "pretend I haven't done anything with it".
	private final String ANIMUNCODED = "anim-uncoded";
	private final String ANTHROUNCODED = "anthro-uncoded";
	private final String ANIMCONFUNCODED = "animconf-uncoded";
	private final String STATUSUNCODED = "status-uncoded";
	private final String STATUSTYPEUNCODED = "statustype-uncoded";

	// a funny symbol to show on screen meaning "uncoded", used
	// animacy in the Stanford coding and status in the Edinburgh coding
	// (since these are the main attributes - without them, the user
	// isn't meant to make choices for the other attributes).
	private final String UNCODED_STRING = "\u00B7";

	// names of buttons for moving forward to the next markable and 
	// back to the previous one
	private final String FORWARD = "forward";
	private final String BACKWARD = "back";

	// set redisplay=false if you want to speed up the interface
	// by coding blind (no longer necessary and may no longer work)
	private final boolean redisplay = true;

	/* STORAGE FOR THE DATA */
	private NOMWriteCorpus nom;
	private NiteMetaData controlData;

	/* STORAGE FOR COMMAND LINE ARGUMENTS */
	private String corpusname;
	private String observationname;
	private String fontsize; //contains a string rep. of a number, like "12" 
	private boolean showsyntax;

	/* COMPONENTS */
	// the frame outside everything
	private JFrame frame;
	// a desktop within that frame, containing the transcription and coding windows
	private JDesktopPane desktop;
	// the component that shows the transcription.
	private JInternalFrame textIFrame;
	// ATTEMPT TO WIRE IN SOUND
	private DefaultClock clock;
	// a tree for showing the links in 
	private NTree ntree;
	// a tree for showing the link type ontology in
	//private NTree lttree;
	// the text area in that frame
	private NTextArea nta;
	// the component that has all of the coding buttons on it for info status.  It itself has
	// two components within it, a panel with the skip button, and the panels 
	// for the various types of codes.
	private JInternalFrame skipIFrame;
	private JInternalFrame noteIFrame; //type-in window for notes.
	private JTextField noteTextField;

	private JInternalFrame infostatusbuttonsIFrame;
	private JInternalFrame statustypebuttonsIFrame;
	private JInternalFrame linkIFrame;
	private JInternalFrame linkactionIFrame;

	// some memory about what's been done with the interface
	private NOMElement changed = null;
	private NOMElement changed_link = null;
	private TreeNode changed_treenode = null;
	private TreePath rootpath = null;
	//private TreePath linktyperootpath=null;
	private NTextElement high = null;

	private String refstr = "anaphor";
	private String antestr = "antecedent";
	private String newlink = "Add new link";
	private String deletelink = "Delete link";
	//    private String newlinktext = "Adds a new link with the highlighted markable as the referent. Assumes the next clicked markable is the antecedent and the next clicked link type is the type.";
	private String newlinktext =
		"Highlighted markable is the anaphor; selected another markable for antecedent";
	private String antecedent = "Add / alter antecedent";
	private String referent = "Add / alter referent";
	//private String linktype = "Change link type";
	//private String linktypeel = "linktype";
	//private String linktypeatt = "type";
	private Color refcolour = Color.lightGray;
	private Color antecolour = Color.pink;

	/* CONTROLLING HOW THE THING LOOKS */
	// fontsize command line argument will determine textsize (the fontsize
	// to use for basic text and for button labels) and bracketsize (the fontsize
	// to use for brackets, slightly larger to make them easier to click on)
	private int textsize;
	private int bracketsize;

	// global counter of markable nesting, so that we can colour them differently
	private int marknest = 0;

	// NITE text areas require one to construct styles in advance to use them.
	// We need two types for this interface:  ones for normal text, and ones
	// for brackets in a slightly larger size.  We'll keep these in two separate
	// lists.
	// The stylenamelists have two purposes:  telling the colour name to use
	// and giving a name that the text area can use to index the corresponding style.
	// Since the names of the styles need to be unique, we'll be appending "text"
	// to the styles for the normal text, and leaving stylenames from the bracket
	// list as they are.  
	private String[] normalstylenamelist = { "black" };
	private Style[] normalstylelist = new Style[normalstylenamelist.length];
	private String[] bracketstylenamelist =
		{ "black", "blue", "red", "green", "yellow" };
	private Style[] bracketstylelist = new Style[bracketstylenamelist.length];

	private Engine searchEngine = new Engine();
	private Stack markable_stack = new Stack();
	private LinkedList markable_list = new LinkedList();
	private JButton newbut;
        private String redball_filename = "/net/sourceforge/nite/icons/misc/ballred.gif";
	private String greenball_filename = "/net/sourceforge/nite/icons/misc/ballgreen.gif";
	private ImageIcon highlight_image;
	private ImageIcon highlight_green_image;
	// these are maps we use to highlight the current code on the buttons
	private HashMap linkmap = new HashMap();
	private HashMap statustypebuttonmap = new HashMap();
	private HashMap statusbuttonmap = new HashMap();
	private boolean WAITING_FOR_ANTECEDENT_MODE = false;
    //private QueryWindow search = null;
    private net.sourceforge.nite.search.GUI search = null;
	///////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////
	public SwitchboardDisplay() {
	}

	public SwitchboardDisplay(
		String c,
		String o,
		String s,
		boolean syn) {
		corpusname = c;
		observationname = o;
		fontsize = s;
		showsyntax = syn;
		try {
			controlData = new NiteMetaData(corpusname);
			nom = new NOMWriteCorpus(controlData);
			if (observationname == null) {
				ChooseObservation co = new ChooseObservation(controlData);
				observationname = co.popupDialog();
			}
			if (observationname != null) {
				NObservation obs =
					controlData.getObservationWithName(observationname);
				nom.loadData(obs);
				System.out.println("Finished loading.");
			}
			//search = new QueryWindow(nom);
			search = new net.sourceforge.nite.search.GUI(nom);
			setupInterface(nom, fontsize);
		} catch (NiteMetaException nme) {
			nme.printStackTrace();
		} catch (NOMException nex) {
			nex.printStackTrace();
		}
	}

	private void setupInterface(NOMWriteCorpus nom, String fontsize) {
		frame = new JFrame();
		frame.addWindowListener(this);
		highlight_image = loadIcon(redball_filename);
		highlight_green_image = loadIcon(greenball_filename);
		frame.setTitle(
			"Switchboard Corpus Display -- "
				+ observationname);
		makeFileMenu(frame);
		desktop = new JDesktopPane();
		figureFontSize(fontsize);
		makeStyles(textsize, bracketsize);
		noteIFrame = new JInternalFrame("Note", true, false, true, true);
		noteIFrame.setSize(new Dimension(500, 100));
		noteIFrame.setLocation(new Point(0, 620));
		makeNoteTypeIn(noteIFrame);
		desktop.add(noteIFrame);

		/* LINKS */
		linkIFrame = new JInternalFrame("Links", true, false, true, true);
		linkIFrame.setSize(new Dimension(300, 245));
		linkIFrame.setLocation(new Point(520, 250));
		addLinkPaneContent(linkIFrame);
		desktop.add(linkIFrame);

		/* SIGNAL */
		//JC:  I'm really surprised I can't figure out how to get the signals 
		//to play in modern NXT.  
		//		DefaultClock clock = new DefaultClock();
		//		clock.ensureVisible(new Rectangle(500,-150, 200, 200), desktop);

		/* MAIN DISPLAY */ 
		textIFrame =
			new JInternalFrame(
				"Transcription Display",
				true,
				false,
				true,
				true);
		textIFrame.setSize(new Dimension(500, 500));
		addTextAreaContent(textIFrame);
		/* Find the element for initial highlight */
		//if (markable_list.size() != 0) {
		//	changed = (NOMElement) markable_list.getFirst();
		//} 
		desktop.add(textIFrame);

		desktop.setSize(new Dimension(900, 750));
		frame.getContentPane().add(desktop);
		frame.setSize(new Dimension(920, 770));
		frame.setVisible(true);

		//if (changed != null) {
		//	nta.setHighlighted(
		//		new NOMObjectModelElement(changed),
		//		MetalLookAndFeel.getTextHighlightColor());
		//	showCodeHighlights(changed);
		//	showNote(changed);
		//}
	}

	/** Turn off the little icons next to coding buttons for
	 * all the sets of buttons that use them on this interface.
	 */
	private void clearCodeHighlights() {
		clearIcons(statusbuttonmap);
		clearIcons(statustypebuttonmap);
	}

	/** Update the little icons next to coding buttons for
	 * all the sets of buttons that use them on this interface.
	 */
	private void showCodeHighlights(NOMElement nomel) {
		showCode(
			nomel,
			STATUS,
			STATUSUNCODED,
			statusbuttonmap,
			highlight_image);
		showCode(
			nomel,
			STATUSTYPE,
			STATUSTYPEUNCODED,
			statustypebuttonmap,
			highlight_green_image);
	}

	private void showNote(NOMElement nomel) {
		String text = (String) nomel.getAttributeComparableValue(NOTE);
		noteTextField.setText(text);
	}

	private void clearNote() {
		noteTextField.setText("");
	}

	/** add a file menu.  Most of these facilities draw on things from the
	 * gui.util package.
	 */
	private void makeFileMenu(JFrame frame) {
		JMenuBar menubar = new JMenuBar();
		JMenu file = new JMenu("File");
		file.add(new OpenObservationAction());
		//file.add(new SaveAction());
		//file.add(new SaveAsAction());
		file.add(new RedisplayAction());
		file.add(new FontSizeAction());
		file.add(new PrintAction());
		file.add(new ExitAction());
		menubar.add(file);
		/* Make the search menu too! */
		JMenu search = new JMenu("Search");
		search.add(new SearchAction());
		menubar.add(search);

		frame.setJMenuBar(menubar);
	}

	/** put up a notes section; when the user hits return, it adds
	 * the note for the currently selected markable.
	 */
	private void makeNoteTypeIn(JInternalFrame f) {
		JPanel panel = new JPanel();
		//JButton nbut = new JButton(NOTEBUTTON);
		//nbut.setFont(new Font("Serif", Font.PLAIN, textsize));
		//nbut.addActionListener(this);
		//panel.add(nbut);
		f.getContentPane().add(panel);
		noteTextField = new JTextField(20);
		//noteTextField.addActionListener(this);
		panel.add(noteTextField);
		f.setVisible(true);

	}

	/** put buttons corresponding to skip forwards and skip backwards
	 *  that don't change the coding of the current markable, put
	 * them on a panel on the internal frame.
	 */
	private void makeSkipAndBackButtons(JInternalFrame f) {
		//JSplitPane jsp = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		//JPanel buttonspanel = new JPanel();
		//addButtons(buttonspanel, elname, attname);
		JPanel skipbuttonpanel = new JPanel();
		addSkipButton(skipbuttonpanel);
		addBackButton(skipbuttonpanel);
		//jsp.setTopComponent(skipbuttonpanel);
		//jsp.setBottomComponent(buttonspanel);
		//	f.getContentPane().add(panel);
		//f.getContentPane().add(jsp);
		f.getContentPane().add(skipbuttonpanel);
		f.setVisible(true);

	}
	/** put buttons corresponding to the enumerated attribute values of the
	* named attribute and element on the internal frame, and add the frame
	* to the desktop
	*/
	private HashMap makeCodingButtons(
		JInternalFrame f,
		String elname,
		String attname) {
		JPanel panel = new JPanel();
		HashMap bmap = addButtons(panel, elname, attname);
		f.getContentPane().add(panel);
		f.setVisible(true);
		return bmap;
	}

	/** Add the buttons to set the chosen attribute for the chosen element,
	 * drawing the values from the metadata description
	 */
	private HashMap addButtons(JPanel panel, String elname, String attname) {
		NElement nell = controlData.getElementByName(elname);
		NAttribute nat = nell.getAttributeByName(attname);
		HashMap bmap = new HashMap();

		List vals = nat.getEnumeratedValues();
		Iterator nit = vals.iterator();
		while (nit.hasNext()) {
			String val = (String) nit.next();
			JButton nbut = new JButton(val);
			nbut.setFont(new Font("Serif", Font.PLAIN, textsize));
			nbut.addActionListener(this);
			bmap.put(val, nbut);
			panel.add(nbut);
		}
		return bmap;
	}

	/** Add an extra button on another panel in the same frame
	 * as the markable coding buttons that skips forward, leaving
	 * the coding exactly as it was.  (The coding interface moves on
	 * to the next markable when one is coded).
	 */
	private void addSkipButton(JPanel panel) {
		JButton nbut = new JButton(FORWARD);
		nbut.setFont(new Font("Serif", Font.PLAIN, textsize));
		nbut.addActionListener(this);
		panel.add(nbut);
	}

	/** Add an extra button on another panel in the same frame
	 * as the markable coding buttons that skips backward, leaving
	 * the coding exactly as it was.  
	 */
	private void addBackButton(JPanel panel) {
		JButton nbut = new JButton(BACKWARD);
		nbut.setFont(new Font("Serif", Font.PLAIN, textsize));
		nbut.addActionListener(this);
		panel.add(nbut);
	}

	/* Add buttons for link actions */
	private void addLinkButtons(JInternalFrame frame) {
		JPanel panel = new JPanel();
		JButton nbut = new JButton(newlink);
		nbut.addActionListener(this);
		nbut.setToolTipText(newlinktext);
		panel.add(nbut);
		nbut = new JButton(deletelink);
		nbut.addActionListener(this);
		panel.add(nbut);
		nbut = new JButton(antecedent);
		nbut.addActionListener(this);
		panel.add(nbut);
		frame.getContentPane().add(panel);
		//	nbut = new JButton(referent);
		//	nbut.addActionListener(this);
		//	panel.add(nbut);
		// don't need button for changing link type
		//nbut = new JButton(linktype);
		//nbut.addActionListener(this);
		//panel.add(nbut);
	}

	/* display the list of links */
	private void addLinkPaneContent(JInternalFrame frame) {
		List elist = null;
		linkmap = new HashMap();
		try {
			elist = searchEngine.search(nom, "($a link)");
		} catch (Throwable e) {
			e.printStackTrace();
			System.exit(0);
		}
		if (elist == null || elist.size() <= 1) {
			System.err.println("NO LINK ELEMENTS IN CORPUS!");
			//	    System.exit(0);
		}

		NTreeNode root = new NTreeNode("Links");

		for (int i = 1; i < elist.size(); i++) {
			List l = (List) elist.get(i);
			showLink(root, (NOMElement) l.get(0));
		}

		ntree = new NTree(root);
		ntree.addTreeSelectionListener(this);
		ntree.setCellRenderer(new NTreeCellRenderer());
		rootpath = new TreePath(root);
		JScrollPane pane = new JScrollPane(ntree);
		frame.getContentPane().add(pane);
		frame.setVisible(true);
		if (changed_treenode != null) {
			// select the previous selection
			TreePath path =
				new TreePath(
					((DefaultTreeModel) ntree.getModel()).getPathToRoot(
						changed_treenode));
			ntree.addSelectionPath(path);
		}
		//	ntree.visitAllNodesForData();
	}

	/* Display a link by adding it to the NTree */
	private void showLink(NTreeNode parent, NOMElement link) {
		String text = link.getID() + " ";
		//NOMPointer np=link.getPointerWithRole(linktypeatt);
		//if (np!=null) {
		//    text += "(" + np.getToElement().getAttribute(linktypeatt).getStringValue() + ")"; 
		//} else {
		//    text += "(uncoded)";
		//}
		NTreeNode newnode = new NTreeNode(text);
		NOMObjectModelElement nome = new NOMObjectModelElement(link);
		newnode.setDataElement((ObjectModelElement) nome);
		if (link.getPointers() != null) {
			for (Iterator pit = link.getPointers().iterator();
				pit.hasNext();
				) {
				NOMPointer pointer = (NOMPointer) pit.next();
				//if (pointer.getRole().equalsIgnoreCase(linktypeatt)) { continue; }
				NTreeNode nnode = new NTreeNode(pointer.getRole());
				if (pointer.getRole().equalsIgnoreCase(refstr)) {
					nnode.textcolour = refcolour;
				}
				if (pointer.getRole().equalsIgnoreCase(antestr)) {
					nnode.textcolour = antecolour;
				}
				nome = new NOMObjectModelElement(pointer.getToElement());
				Set linklist = (Set) linkmap.get(pointer.getToElement());
				if (linklist == null) {
					linklist = (Set) new HashSet();
				}
				linklist.add(nome);
				linkmap.put(pointer.getToElement(), linklist);
				nnode.setDataElement((ObjectModelElement) nome);
				newnode.addNode(nnode);
			}
		}
		parent.addNode(newnode);
		if (link == changed_link) {
			changed_treenode = newnode;
		}
	}

	/** Clear a display element frame ready for redisplay 
	 */
	private void clearFrame(JInternalFrame frame) {
		frame.setVisible(false);
		frame.getContentPane().removeAll();
	}

	/** translate the fontsize command line argument into integers for the
	 * font size of the text and buttons, and another, a little larger, for the
	 * brackets, storing in the globals textsize and bracketsize.
	 */
	private void figureFontSize(String fontsize) {
		// before making styles, figure out what size of normal text to use, with 12 as the default.
		textsize = 12;
		if (fontsize != null) {
			try {
				Integer fs = Integer.valueOf(fontsize);
				textsize = fs.intValue();
			} catch (NumberFormatException e) {
				System.out.println(
					"Can't interpret "
						+ fontsize
						+ " as an integer to use as font size!");
				e.printStackTrace();
				System.exit(0);
			}
		}
		bracketsize = textsize;
	}

	/** NITE expects styles for use in a text area to be made ahead.  Iterating down the colour names given in bracketstylenamelist
	 * and textstylenamelist, make the styles and store them in bracketstylelist and textstylelist.  
	 * Styles are made using a TextStyleHandler.  
	 */
	private void makeStyles(int textsize, int bracketsize) {

		for (int i = 0; i < bracketstylenamelist.length; i++) {
			TextStyleHandler stylehandler = new TextStyleHandler();
			stylehandler.init("", null);
			stylehandler.setName(bracketstylenamelist[i]);
			stylehandler.makeNewStyle();
			// args are style, foreground colour, background colour
			stylehandler.setStyle(
				stylehandler.getColours(
					stylehandler.getStyle(),
					bracketstylenamelist[i],
					bracketstylenamelist[i]));

			StyleConstants.setFontSize(stylehandler.getStyle(), bracketsize);
			StyleConstants.setBold(stylehandler.getStyle(), true);
			bracketstylelist[i] = stylehandler.getStyle();
		}
		for (int i = 0; i < normalstylenamelist.length; i++) {
			TextStyleHandler stylehandler = new TextStyleHandler();
			stylehandler.init("", null);
			stylehandler.setName(normalstylenamelist[i] + "text");
			stylehandler.makeNewStyle();
			stylehandler.setStyle(
				stylehandler.getColours(
					stylehandler.getStyle(),
					normalstylenamelist[i],
					normalstylenamelist[i]));
			StyleConstants.setFontSize(stylehandler.getStyle(), textsize);
			StyleConstants.setBold(stylehandler.getStyle(), false);
			normalstylelist[i] = stylehandler.getStyle();
		}
	}

	/* display or redisplay the main text area */
	private void addTextAreaContent(JInternalFrame frame) {
		if (nta != null && search != null) {
			search.deregisterResultHandler(nta);
		}
		nta = new NTextArea();
		if (search != null) {
			search.registerResultHandler(nta);
		}
		nta.addMouseListener(this);
		// add the styles to the text area
		for (int i = 0; i < bracketstylenamelist.length; i++) {
			nta.addStyle(bracketstylenamelist[i], bracketstylelist[i]);
		}
		for (int i = 0; i < normalstylenamelist.length; i++) {
			nta.addStyle(normalstylenamelist[i] + "text", normalstylelist[i]);
		}

		// create a list of all the parse elements in the loaded data.
		List elist = null;
		try {
			elist = searchEngine.search(nom, "($a parse)");
		} catch (Throwable e) {
			e.printStackTrace();
			System.exit(0);
		}
		if (elist == null || elist.size() <= 1) {
			//	    System.err.println("NO parse ELEMENTS IN CORPUS!");
			//	    System.exit(0);
		}

		/* The first thing on the list returned by the search engine
		 * is an entry containing the names of the variables, so start
		 * iterating down the list from 1
		 */
		for (int i = 1; i < elist.size(); i++) {
			List l = (List) elist.get(i);
			NOMElement nwe = (NOMElement) l.get(0);
			// show something that uniquely identifies the parse (like the
			// id), plus the speaker.
			showParseField(nwe);
			//render the words in the parse
			showWords(nwe, 0);
			// render two newlines
   			   NTextElement nte = new NTextElement("\n\n", "blacktext");
			   NOMObjectModelElement nome = new NOMObjectModelElement(nwe);
			   nte.setDataElement((ObjectModelElement) nome);
			   nta.addElement(nte);
		}

		JScrollPane pane = new JScrollPane(nta);
		frame.getContentPane().add(pane);
		frame.setVisible(true);
	}

	/** given a parse element, pack a text element in nta
	 * that shows the speaker of the turn parent, hyphen,
	 * parse id.
	 */
	private void showParseField(NOMElement nwe) {
		if (nwe == null) {
			return;
		}
		//figure out what the speaker is
		String speaker = null;
		speaker=nwe.getAgentName();
		//get the id
		String id = nwe.getID();
		if ((speaker == null) || (id == null)) {
			System.out.println(
				"Found a parse missing either a speaker or an id!");
		} else {
			// canonical way of adding a text element to an existing 
			// text area
			//create the text element with the desired string and style
			String content = speaker + "-" + id + ": ";
			if (showsyntax) {
				content += "\n";
			};
			NTextElement nte = new NTextElement(content, "red");
			//create a NOMObjectModelElement to relate data to display
			NOMObjectModelElement nome = new NOMObjectModelElement(nwe);
			//do the relating
			nte.setDataElement((ObjectModelElement) nome);
			//add the text element to the text area
			nta.addElement(nte);
			// see if we can attach a note to it
			ChangeAttributeValueOptionPane op =
				new ChangeAttributeValueOptionPane(nome, "Enter note:", "note");
		}
	}

    private String levelString (int level) {
	String retval = "";
	if (showsyntax) {
          for (int i=1;i<=(level*10);++i) {
	    retval += " ";
	  };
	} else {
            retval = " ";
	};
	return retval;
    }

    /** show the words in the corpus, recursively descending through
	the syntax layers and checking for markables at each stage. */
	private void showWords(NOMElement nwe, int level) {
            String endline = "";
	    if (showsyntax) {
		endline="\n";
	    }
		if (nwe == null || nwe.getChildren() == null) {
			return;
		}
		showMarkableParentsStart(nwe, level);
		for (Iterator kit = nwe.getChildren().iterator(); kit.hasNext();) {
			NOMElement kid = (NOMElement) kit.next();
			if (kid.getName().equalsIgnoreCase("word")) {
				showMarkableParentsStart(kid, level);
				String text = null;
				try {
				    text = (String) kid.getAttributeComparableValue("orth");
				    if (showsyntax) {
					text += "-";
					text += (String) kid.getAttributeComparableValue("pos");
				    }
				} catch (Exception e) {
				};
				NOMObjectModelElement nome = new NOMObjectModelElement(kid);
				if (!markable_stack.empty()) {
					NOMObjectModelElement parnome =
						(NOMObjectModelElement) markable_stack.peek();
					nta.addElement(
						new NTextElement(
							endline + levelString(level) + text + " " ,
							"blacktext",
							nome,
							parnome));
				} else {
					nta.addElement(
						new NTextElement(
							endline + levelString(level) + text + " ",
							"blacktext",
							nome));
				}
				showMarkableParentsEnd(kid, level);
			} else if (kid.getName().equalsIgnoreCase("trace")) {
				showMarkableParentsStart(kid,level);
				NOMObjectModelElement nome = new NOMObjectModelElement(kid);
				//instead of just saying <trace>, put angle brackets around the label attribute
				//of the movement that has this trace filling its target role
				String tracetype = null;
				List pointerlist = kid.getPointersTo();
				if (pointerlist == null) {
					//System.out.println(
					//	"Found a trace with nothing that points to it!");
				} else {
					for (Iterator pit = kid.getPointersTo().iterator();
						pit.hasNext();
						) {
						NOMPointer ptr = (NOMPointer) pit.next();
						if (ptr.getRole().equalsIgnoreCase("target")) {
							tracetype =
								(String) ptr
									.getFromElement()
									.getAttributeComparableValue(
									"label");
						}
					};
				};
				if (tracetype == null) {
					//System.out.println(
					//	"Found a trace missing either a target role or the label of the pointing element!");
					tracetype = "trace";
				};
				if (!markable_stack.empty()) {

					NOMObjectModelElement parnome =
						(NOMObjectModelElement) markable_stack.peek();
					nta.addElement(
						new NTextElement(
							endline + levelString(level) + "<" + tracetype + "> ",
							"blacktext",
							nome,
							parnome));
				} else {
					nta.addElement(
						new NTextElement(
							endline + levelString(level) + "<" + tracetype + "> ",
							"blacktext",
							nome));
				}
				showMarkableParentsEnd(kid,level);
			} else if ((showsyntax) && (kid.getName().equalsIgnoreCase("nt"))) {
				showMarkableParentsStart(kid,level);
				NOMObjectModelElement nome = new NOMObjectModelElement(kid);
				String cat = (String) kid.getAttributeComparableValue("cat");
				String subcat =
					(String) kid.getAttributeComparableValue("subcat");
				String render = cat;
				if (subcat != null) {
					render += "-" + subcat;
				}
				if (!markable_stack.empty()) {

					NOMObjectModelElement parnome =
						(NOMObjectModelElement) markable_stack.peek();
					nta.addElement(
						new NTextElement(
							endline + levelString(level) + "<" + render + "> ",
							"blacktext",
							nome,
							parnome));
				} else {
					nta.addElement(
						new NTextElement(
							endline + levelString(level) + "<" + render + "> ",
							"blacktext",
							nome));
				}
				showWords(kid, level+1);
				showMarkableParentsEnd(kid,level);
			} else {
				showWords(kid, level);
			}
		}
		showMarkableParentsEnd(nwe, level);
	}

	/* display any open-markable elements in the appropriate colour */
	private void showMarkableParentsStart(NOMElement nwe, int level) {
		if (nwe.getPointersTo() == null)
			return;
		for (Iterator kit = nwe.getPointersTo().iterator(); kit.hasNext();) {
			NOMPointer point = (NOMPointer) kit.next();
			if (point == null)
				continue;
			NOMElement parent = point.getFromElement();
			if (parent.getName().equalsIgnoreCase(MARKABLE)) {
				marknest++;
				//String status = "";
				//if (parent.getAttribute(STATUS)!=null) {
				// previously, checker added space, nonchecker didn't
				//    status=parent.getAttribute(STATUS).getStringValue();
				//}
				int index = marknest;
				if (marknest >= bracketstylenamelist.length) {
					index = (marknest % bracketstylenamelist.length) + 1;
					if (index == bracketstylenamelist.length) {
						index = 1;
					}
				}
				String sname = bracketstylenamelist[index];
				NTextElement nte =
					new NTextElement(statusStartString(parent, level), sname);
				NOMObjectModelElement nome = new NOMObjectModelElement(parent);
				nome.setDisplayedAttribute(STATUS);
				nte.setDataElement((ObjectModelElement) nome);
				nta.addElement(nte);
				markable_stack.push(nome);
				markable_list.add(parent);
				//if (parent == changed) {
				//	high = nte;
				//}
			}
		}
	}

	/* display any close-markable elements in the appropriate colour */
	private void showMarkableParentsEnd(NOMElement nwe, int level) {
		if (nwe.getPointersTo() == null)
			return;
		for (Iterator kit = nwe.getPointersTo().iterator(); kit.hasNext();) {
			NOMPointer point = (NOMPointer) kit.next();
			if (point == null)
				continue;
			NOMElement parent = point.getFromElement();
			if (parent.getName().equalsIgnoreCase(MARKABLE)) {
				int index = marknest;
				if (marknest >= bracketstylenamelist.length) {
					index = (marknest % bracketstylenamelist.length) + 1;
					if (index == bracketstylenamelist.length) {
						index = 1;
					}
				}
				String sname = bracketstylenamelist[index];
				//String status = "";
				//if (parent.getAttribute(STATUS)!=null) {
				//    status=parent.getAttribute(STATUS).getStringValue() + " ";
				//}
				NTextElement nte =
					new NTextElement(statusEndString(parent, level), sname);
				NOMObjectModelElement nome = new NOMObjectModelElement(parent);
				nte.setDataElement((ObjectModelElement) nome);
				nta.addElement(nte);
				nte = new NTextElement(" ", null);
				nta.addElement(nte);
				markable_stack.pop();
				marknest--;
			}
		}
	}


	/* this is where we could select special chars to begin certain
	   types of markable */
	private String statusStartString(NOMElement el, int level) {
		String retstr = "";
		if (showsyntax) {
			retstr += "\n" + levelString(level);
		}
		retstr += "(";
		String status = null;
		if (el.getAttribute(STATUS) != null) {
		    retstr += el.getAttribute(STATUS).getStringValue();
		};
		if (el.getAttribute(STATUSTYPE) != null) {
		    retstr += "/" + el.getAttribute(STATUSTYPE).getStringValue();
		};
		if (el.getAttribute(ANIMACY) != null) {
		    retstr += "//" + el.getAttribute(ANIMACY).getStringValue();
		};
		if (el.getAttribute(ANIMCONF) != null) {
		    if (el
			.getAttribute(ANIMCONF)
			.getStringValue()
			.equals("vanim")) {
			retstr += "/v";
		    }
		};
		if (el.getAttribute(ANTHRO) != null) {
		    if (el
			.getAttribute(ANTHRO)
			.getStringValue()
			.equals("anthro")) {
			retstr += "/a";
		    }
		}
		return retstr;
	}

	/* this is where we could select special chars to end certain
	   types of markable */
	private String statusEndString(NOMElement el,int level) {
	    String retstr = "";
		if (showsyntax) {
  		    retstr += "\n";
		}
  		retstr += levelString(level) + ")";
		return retstr;
	}

	/** TreeSelectionEvents - highlight the referent and antecedent on
	 the text area */
	public void valueChanged(TreeSelectionEvent tse) {
		//if (linktyperootpath.isDescendant(tse.getNewLeadSelectionPath())) {
		//    linkTypeSelected(tse);
		//   return;
		//}
		NOMElement nel = findLink();
		WAITING_FOR_ANTECEDENT_MODE = false;
		if (nel == null) {
			return;
		}
		nta.clearHighlights(NTextArea.SELECTION_HIGHLIGHTS);
		//clearCodeHighlights();
		if (nel.getName().equalsIgnoreCase("link")) {
			for (Iterator pit = nel.getPointers().iterator(); pit.hasNext();) {
				NOMPointer pointer = (NOMPointer) pit.next();
				if (pointer.getRole().equalsIgnoreCase(antestr)) {
					highlightAntecedent(pointer.getToElement());
				} else if (pointer.getRole().equalsIgnoreCase(refstr)) {
					highlightReferent(pointer.getToElement());
				}
			}
			changed_link = nel;
		} else {
			// this is a bit cheesy - testing the string because we
			// can't associate the Pointer with the text element
			String path =
				((TreeNode) tse
					.getNewLeadSelectionPath()
					.getLastPathComponent())
					.toString();
			if (path.indexOf(refstr) >= 0) {
				highlightReferent(nel);
			} else if (path.indexOf(antestr) >= 0) {
				highlightAntecedent(nel);
			} else {
			    nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(nel));
			}
		}
	}

	private void highlightAntecedent(NOMElement ante) {
	    nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(ante));
	}

	private void highlightReferent(NOMElement ref) {
	    nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(ref));
	}

	/*
	private void linkTypeSelected(TreeSelectionEvent tse) {
	//	System.out.println("Link type selected");
	changeLinkType();
	}
	*/

	/** Start the application. */
	public static void main(String[] args) {
		String corpus = null;
		String observation = null;
		String size = null;
		boolean syntax = false;
		if (args.length < 2 || args.length > 8) {
			usage();
		}
		for (int i = 0; i < args.length; i++) {
			String flag = args[i];
			if (flag.equals("-corpus") || flag.equals("-c")) {
				i++;
				if (i >= args.length) {
					usage();
				}
				corpus = args[i];
			} else if (flag.equals("-observation") || flag.equals("-o")) {
				i++;
				if (i >= args.length) {
					usage();
				}
				observation = args[i];
			} else if (flag.equals("-fontsize") || flag.equals("-f")) {
				i++;
				if (i >= args.length) {
					usage();
				}
				size = args[i];
			} else if (flag.equals("-syntax") || flag.equals("-s")) {
				syntax = true;
			} else {
				usage();
			}
		}
		if (corpus == null) {
			usage();
		}

		SwitchboardDisplay m =
			new SwitchboardDisplay(corpus, observation, size, syntax);
	}

	private static void usage() {
		System.err.println(
			"Usage: java SwitchboardDisplay -c metadata-filename -o observation-name -fontsize size -syntax");
		System.exit(0);
	}

	/** This handles all the button-presses etc for the interface */
	public void actionPerformed(ActionEvent ae) {
		String act = ae.getActionCommand();
		//clearCodeHighlights();
		Set sel = nta.getSelectedElements();
		if (sel == null || sel.size() == 0) {
			System.out.println("Nothing selected in text area");
		} else {
			Iterator sit = sel.iterator();
			while (sit.hasNext()) {
				NOMObjectModelElement el = (NOMObjectModelElement) sit.next();
				if (el == null) {
					continue;
				}
				if (!el.getName().equalsIgnoreCase(MARKABLE)) {
					System.out.println("Please select a markable");
					return;
				}				
		}

		nta.clearHighlights(NTextArea.SELECTION_HIGHLIGHTS);
		//nta.setHighlighted(
		//		   new NOMObjectModelElement(changed),
		//	MetalLookAndFeel.getTextHighlightColor());
		//showCodeHighlights(changed);
		//showNote(changed);
		}
	}





	/** find the highlighted link */
	private NOMWriteElement findLink() {
		Set sel = ntree.getSelectedTreeNodes();
		if (sel == null
			|| sel.iterator() == null
			|| !sel.iterator().hasNext()) {
			System.out.println("Please select a link!");
			return null;
		} else {
			NTreeNode ntn = (NTreeNode) sel.iterator().next();
			NOMObjectModelElement el =
				(NOMObjectModelElement) ntn.getDataElement();
			if (el == null) {
				return null;
			}
			return (NOMWriteElement) el.getElement();
		}
	}

	/** find the highlighted link type
	private NOMWriteElement findLinkType() {
	Set sel = lttree.getSelectedTreeNodes();
	if (sel == null || sel.iterator()==null || !sel.iterator().hasNext()) {
	    System.out.println("Please select a link type!");
	    return null;
	} else {
	    NTreeNode ntn = (NTreeNode) sel.iterator().next();
	    NOMObjectModelElement el = (NOMObjectModelElement) ntn.getDataElement();
	    if (el==null) { return null; }
	    return (NOMWriteElement)el.getElement();
	}
	}
	*/

	/** Save the NOM using the parameters set on the interface
	 */
	public class SaveAction extends AbstractAction {
		public SaveAction() {
			super("Save observation");
		}
		public void actionPerformed(ActionEvent ev) {
			try {
				controlData.writeMetaData(corpusname);
				nom.serializeCorpusChanged();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	/** Save the changes to the corpus using the dialogue provided   */
	public class SaveAsAction extends AbstractAction {
		public SaveAsAction() {
			super("Save as...");
		}
		public void actionPerformed(ActionEvent ev) {
			CorpusSaveOptions cso = new CorpusSaveOptions(nom);
		}
	}

	/** Redisplay (in case the screen gets messed up)  */
	public class RedisplayAction extends AbstractAction {
		public RedisplayAction() {
			super("Redisplay...");
		}
		public void actionPerformed(ActionEvent ev) {
			Set sel = nta.getSelectedElements();
			if (sel != null || sel.size() != 0) {
				NOMObjectModelElement el =
					(NOMObjectModelElement) sel.iterator().next();
				changed = el.getElement();
			}
			clearFrame(textIFrame);
			addTextAreaContent(textIFrame);
			nta.clearHighlights(NTextArea.SELECTION_HIGHLIGHTS);
			//if (changed != null) {
			//	nta.setHighlighted(
			//		new NOMObjectModelElement(changed),
			//		MetalLookAndFeel.getTextHighlightColor());
			//	showCodeHighlights(changed);
			//	showNote(changed);
			//}
		}
	}

	/** change the font size (which also changes the size of the buttons) */
	public class FontSizeAction extends AbstractAction {
		public FontSizeAction() {
			super("Font Size...");
		}
		public void actionPerformed(ActionEvent ev) {
			ChooseFontSize cfs = new ChooseFontSize(textsize);
			String fontsize = cfs.popupDialog();

			// I think we can just create new styles overwriting the old names,
			// but if not, need code here that nta.removeStyle(Style) for each one.
			figureFontSize(fontsize);
			makeStyles(textsize, bracketsize);
			clearFrame(infostatusbuttonsIFrame);
			statusbuttonmap =
				makeCodingButtons(infostatusbuttonsIFrame, MARKABLE, STATUS);

			clearFrame(statustypebuttonsIFrame);
			statustypebuttonmap =
				makeCodingButtons(
					statustypebuttonsIFrame,
					MARKABLE,
					STATUSTYPE);

			clearFrame(skipIFrame);
			makeSkipAndBackButtons(skipIFrame);

			clearFrame(noteIFrame);
			makeNoteTypeIn(noteIFrame);

			//clearFrame(linkIFrame);
			//addLinkPaneContent(linkIFrame);

			//clearFrame(linkactionIFrame);
			//addLinkButtons(linkactionIFrame);

			//redisplay main text area
			Set sel = nta.getSelectedElements();
			//if (sel != null || sel.size() != 0) {
			//	NOMObjectModelElement el =
			//		(NOMObjectModelElement) sel.iterator().next();
			//	changed = el.getElement();
			//}
			clearFrame(textIFrame);
			addTextAreaContent(textIFrame);
			nta.clearHighlights(NTextArea.SELECTION_HIGHLIGHTS);
			//if (changed != null) {
			//	nta.setHighlighted(
			//		new NOMObjectModelElement(changed),
			//		MetalLookAndFeel.getTextHighlightColor());
			//	showCodeHighlights(changed);
			//	showNote(changed);
			//}
		}
	}

	/** A very simple "exit" action - just checks for changes and saves if so */
	public class ExitAction extends AbstractAction {
		public ExitAction() {
			super("Exit");
		}
		public void actionPerformed(ActionEvent ev) {
			exit();
		}
	}

	private void exit() {
		CheckSave co = new CheckSave(nom);
		if (search != null) {
			search.savePreferences();
		}
		if (!(co.popupDialog() == JOptionPane.CANCEL_OPTION)) {
			System.exit(0);
		}
	}

	/** A "search" action - pops up a search GUI */
	public class SearchAction extends AbstractAction {
		public SearchAction() {
			super("Search...");
		}
		public void actionPerformed(ActionEvent ev) {
			search.popupSearchWindow();
		}
	}

	/** Change observation but check for changes first */
	public class OpenObservationAction extends AbstractAction {
		public OpenObservationAction() {
			super("Open...");
		}
		public void actionPerformed(ActionEvent ev) {
			CheckSave co = new CheckSave(nom);
			if (!(co.popupDialog() == JOptionPane.CANCEL_OPTION)) {
				ChooseObservation cob = new ChooseObservation(controlData);
				String ob = cob.popupDialog();
				if (ob == null || ob.length() == 0)
					return;
				observationname = ob;
				System.out.println("Opening " + observationname);
				nom.clearData();
				markable_list = new LinkedList();
				NObservation obs =
					controlData.getObservationWithName(observationname);
				try {
					nom.loadData(obs);
				} catch (NOMException nex) {
					nex.printStackTrace();
				}
				clearFrame(textIFrame);
				addTextAreaContent(textIFrame);
				nta.clearHighlights(NTextArea.SELECTION_HIGHLIGHTS);
				changed = (NOMElement) markable_list.getFirst();
				nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(changed));
				clearCodeHighlights();
			}
		}
	}

	/** Print text area */
	public class PrintAction extends AbstractAction {
		public PrintAction() {
			super("Print...");
		}
		public void actionPerformed(ActionEvent ev) {
			nta.printTextArea();
		}
	}

	private void clearIcons(HashMap bmap) {
		for (Iterator bit = bmap.values().iterator(); bit.hasNext();) {
			JButton but = (JButton) bit.next();
			if (but.getIcon() != null) {
				but.setIcon(null);
				but.repaint();
			}
		}
	}

	private void clearLinks() {
		ntree.clearHighlights();
	}

	/** paint an image icon on the button corresponding to the attribute for the 
	 * current element, unless it's the uncoded value that's been made explicit
	 */
	private void showCode(
		NOMElement nel,
		String att,
		String uncodedval,
		HashMap bmap,
		ImageIcon im) {
		//System.out.println("Showing code for " + nel.getName() + " with " + att);
		String st = (String) nel.getAttributeComparableValue(att);
		if ((st == null) || (st == uncodedval)) {
			return;
		}
		JButton mybutton = null;
		if (bmap.get(st) != null) {
			mybutton = (JButton) bmap.get(st);
		}
		if (mybutton != null) {
			mybutton.setIcon(im);
			mybutton.repaint();
		}
	}

	private void showLinks(NOMElement nel) {
		if (linkmap.get(nel) == null) {
			return;
		}
		Set links = (Set) linkmap.get(nel);
		//	System.out.println("Iterate: " + nel.getID());
		for (Iterator lit = links.iterator(); lit.hasNext();) {
			ObjectModelElement lel = (ObjectModelElement) lit.next();
			ntree.setHighlighted(lel, antecolour);
		}
	}

	/* Mouse listener implementation */
	public void mouseClicked(MouseEvent e) {

	}
	public void mouseEntered(MouseEvent e) {

	}
	public void mouseExited(MouseEvent e) {

	}
	public void mousePressed(MouseEvent e) {

	}

	public void mouseReleased(MouseEvent e) {
		//	System.out.println("Release");
		Set sel = nta.getSelectedElements();
		if (sel == null || sel.size() == 0) {
			System.out.println("Nothing selected in text area");
			return;
		}
		//clearCodeHighlights();
		clearNote();
		clearLinks();
		for (Iterator sit = sel.iterator(); sit.hasNext();) {
			NOMObjectModelElement el = (NOMObjectModelElement) sit.next();
			if (el.getName().equals(MARKABLE)) {
				//System.out.println("Selection: " + el.getElement().getID());
				NOMElement nel = el.getElement();
				//showCodeHighlights(nel);
				showNote(nel);
				showLinks(nel);
			}
		}
	}

	private ImageIcon loadIcon(String name) {
	    try {
		// get resource
		URL resource = ClassLoader.getSystemClassLoader().getResource(name);
		if (resource == null) {
		    resource = getClass().getResource(name);
		}
		if( resource == null ){ 
		    System.err.println("WARNING: Failed to load image " + name);
		    return null;
		}
		return new ImageIcon(frame.getToolkit().getImage(resource));
	    } catch (Exception e) {
		System.err.println("WARNING: Failed to load image " + name);
		return null;
	    }
	}

	/////////////////////////////////////////////////////////////////////////////
	// WindowListener

	/** NOP (method needed for implementing WindowListener).
	 * @param event parameter isn't used */
	public void windowActivated(WindowEvent event) {
	}

	/** NOP (method needed for implementing WindowListener).
	 * @param event parameter isn't used */
	public void windowClosed(WindowEvent event) {
	}

	/**
	 * Processes {@linkplain #exit()}.
	 * @param event parameter isn't used
	 */
	public void windowClosing(WindowEvent event) {
		exit();
	}

	/** NOP (method needed for implementing WindowListener).
	 * @param event parameter isn't used */
	public void windowDeactivated(WindowEvent event) {
	}

	/** NOP (method needed for implementing WindowListener).
	 * @param event parameter isn't used */
	public void windowDeiconified(WindowEvent event) {
	}

	/** NOP (method needed for implementing WindowListener).
	 * @param event parameter isn't used */
	public void windowIconified(WindowEvent event) {
	}

	/** NOP (method needed for implementing WindowListener).
	 * @param event parameter isn't used */
	public void windowOpened(WindowEvent event) {
	}

}
