import java.util.*;
import java.net.URL;
import java.io.File;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Point;
import java.awt.event.*;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.Highlighter;

import net.sourceforge.nite.search.*;
import net.sourceforge.nite.meta.*;
import net.sourceforge.nite.meta.impl.*;
import net.sourceforge.nite.nom.link.*;
import net.sourceforge.nite.nom.nomwrite.*;
import net.sourceforge.nite.nom.nomwrite.impl.*;
import net.sourceforge.nite.util.SearchResultTimeComparator;
import net.sourceforge.nite.gui.util.*;
import net.sourceforge.nite.gui.textviewer.*;
import net.sourceforge.nite.gui.mediaviewer.*;
import net.sourceforge.nite.nxt.*;
import net.sourceforge.nite.nstyle.handler.*;
import net.sourceforge.nite.time.*;
import net.sourceforge.nite.tools.videolabeler.GlobalInputMap;

import net.sourceforge.nite.nom.NOMException;
import javax.swing.plaf.metal.*;

/**
 * Switchboard interface for markable coding
 * 
 * Jan 08, being dusted off (but nothing more!) for public release of
 * data; name change from SwitchboardEdin to SwitchboardInfoStatus;
 * find text of terminals using orth attribute.
 *
 * Aug 09, a data representation change meant that the transcription no
 * longer interleaved words from different speakers, because the tool
 * was rendering one line per parse, and these are untimed.  I've changed
 * this to rendering one line per turn. (JC)
 *
 * @author Jonathan Kilgour April 2003
 **/

public class SwitchboardKontrast 
    implements ActionListener, TreeSelectionListener, MouseListener {
    private Clock niteclock;
    private final String UNCODED_STRING="\u00B7";
    private final String MARKABLE="kontrast";
    //private final String MARKABLE="markable";
    private final String TRIGGER="trigger";
    private final String STATUS="type";
    private final String STATUSTYPE="statustype";
    // names of buttons for moving forward to the next markable and 
    // back to the previous one
    private final String FORWARD = "forward";
    private final String BACKWARD = "back";
    private final String NOTEBUTTON = "Add note";
    private NOMWriteCorpus nom;
    private NiteMetaData controlData;
    private JInternalFrame iframe;
    private JInternalFrame iframe3;
    private JInternalFrame skipIFrame;
    private JInternalFrame noteIFrame; //type-in window for notes.
    private JTextField noteTextField;
    private String corpusname;
    private String observationname;
    private JFrame frame;
    private String exportdir = ".";
    private TextAreaHandler tah;
    private NTextArea nta;
    private NTree ntree;
    private NTree lttree;
    private JDesktopPane desktop;
    private NOMElement changed=null;
    private NOMElement changed_link=null;
    private TreeNode changed_treenode=null;
    private TreePath rootpath=null;
    private TreePath linktyperootpath=null;
    private NTextElement high=null;
    /*
    private String refstr = "anaphor";
    private String antestr = "antecedent";
    */
    private String refstr = "referent";
    private String antestr = "trigger";
    private String newlink = "Add new link";
    private String deletelink = "Delete link";
    //    private String newlinktext = "Adds a new link with the highlighted markable as the referent. Assumes the next clicked markable is the antecedent and the next clicked link type is the type.";
    private String newlinktext = "Highlighted markable is the anaphor; selected another markable for antecedent";
    private String antecedent = "Add / alter trigger";
    private String referent = "Add / alter referent";
    private String linktype = "Change link type";
    private String linktypeel = "linktype";
    private String linktypeatt = "type";
    private int marknest=0; // global counter of markable nesting
    private String [] stylenamelist = {"black", "blue", "red", "green", "yellow"};
    // names of transcription elements
    private final String WORD="word";
    private final String PUNC="punc";
    private final String SIL="sil";
    private final String ORTH="orth";

    private Style [] stylelist=new Style[stylenamelist.length];
    private Color refcolour = Color.lightGray;
    private Color antecolour = Color.pink;

    private int textsize=12;
    private Engine searchEngine = new Engine();
    private Stack markable_stack = new Stack();
    private LinkedList markable_list = new LinkedList();
    private JButton newbut;
    private String redball_filename = "/net/sourceforge/nite/icons/misc/ballred.gif";
    private String greenball_filename = "/net/sourceforge/nite/icons/misc/ballgreen.gif";
    private ImageIcon highlight_image;
    private ImageIcon highlight_green_image;
    private HashMap linkmap = new HashMap();
    private HashMap buttonmap = new HashMap();
    private HashMap sbuttonmap = new HashMap();
    private boolean WAITING_FOR_ANTECEDENT_MODE=false;
    private net.sourceforge.nite.search.GUI search=null;
    private GlobalInputMap map=GlobalInputMap.getInstance();
    private Comparator mycomp = new SearchResultTimeComparator();
    
    public SwitchboardKontrast(String c, String o) {
	corpusname = c;
	observationname = o;
	try {
	    controlData = new NiteMetaData(corpusname);
	    nom = new NOMWriteCorpus(controlData);
	    nom.setLazyLoading(false);
	    if (observationname==null) {
		ChooseObservation co = new ChooseObservation(controlData);
		observationname=co.popupDialog();
	    } 
	    if (observationname!=null) {
		NObservation obs = controlData.getObservationWithName(observationname);
		nom.loadData(obs);
		System.out.println("Finished loading.");
	    }
	    niteclock = new DefaultClock(controlData, observationname);
	    setupInterface(nom);
	    search=new net.sourceforge.nite.search.GUI(nom);
	    search.registerResultHandler(nta);
	    niteclock.registerTimeHandler(nta);
	} catch (NiteMetaException nme) {
	    nme.printStackTrace();
	} catch (NOMException nex) {
	    nex.printStackTrace();
	}
    }

    private void setupInterface(NOMWriteCorpus nom) {
	frame = new JFrame();
	highlight_image = loadIcon(redball_filename);
	highlight_green_image = loadIcon(greenball_filename);
        frame.setTitle("Switchboard Corpus Markable Coder");
	JMenuBar menubar = new JMenuBar();
	JMenu file = new JMenu("File");
	file.add(new OpenObservationAction());
	file.add(new SaveAction());
	file.add(new SaveAsAction());
	file.add(new RedisplayAction());
	file.add(new PrintAction());
	file.add(new ExitAction());
	menubar.add(file);
	/* Make the search menu too! */
	JMenu search = new JMenu("Search");
	search.add(new SearchAction());
	menubar.add(search);
	frame.setJMenuBar(menubar);
	desktop = new JDesktopPane();
	makeStyles();

	/* STATUS BUTTONS */
	JInternalFrame iframe2 =
	    new JInternalFrame("Kontrast type", true, false, true, true);
	iframe2.setSize(new Dimension(500, 155));
	iframe2.setLocation(new Point(0, 570));
	JPanel panel = new JPanel();
	addButtons(panel);
	iframe2.getContentPane().add(panel);
	iframe2.setVisible(true);
	desktop.add(iframe2);

	skipIFrame = new JInternalFrame("Skip and back buttons", true, false, true, true);
	skipIFrame.setSize(new Dimension(300, 55));
	skipIFrame.setLocation(new Point(500, 570));
	makeSkipAndBackButtons(skipIFrame);
	desktop.add(skipIFrame);

	/* STATUS TYPE BUTTONS */
	JInternalFrame iframe6 =
	    new JInternalFrame("Markable status type", true, false, true, true);
	iframe6.setSize(new Dimension(200, 95));
	iframe6.setLocation(new Point(300, 570));
	JPanel panel2 = new JPanel();
	//addStatusTypeButtons(panel2);
	iframe6.getContentPane().add(panel2);
	iframe6.setVisible(true);
	//desktop.add(iframe6);

	/* LINKS */
	iframe3 = new JInternalFrame("Links", true, false, true, true);
	iframe3.setSize(new Dimension(300, 245));
	iframe3.setLocation(new Point(500, 0));
 	addLinkPaneContent(iframe3);
	desktop.add(iframe3);

	/* LINK ACTIONS */
	JInternalFrame iframe4 =
	    new JInternalFrame("Link Actions", true, false, true, true);
	iframe4.setSize(new Dimension(300, 125));
	iframe4.setLocation(new Point(500, 245));
	panel = new JPanel();
	addLinkButtons(panel);
	iframe4.getContentPane().add(panel);
	iframe4.setVisible(true);
	desktop.add(iframe4);

	/* LINK TYPES */
	JInternalFrame iframe5 =
	    new JInternalFrame("Link type hierarchy", true, false, true, true);
	iframe5.setSize(new Dimension(300, 295));
	iframe5.setLocation(new Point(500, 370));
	//lttree=showOntologyTree(linktypeel, linktypeatt);
	JScrollPane pane = new JScrollPane(lttree);
	//lttree.addTreeSelectionListener(this);
	iframe5.getContentPane().add(pane);
	iframe5.setVisible(true);
	//desktop.add(iframe5);

	/* MARKABLE DISPLAY */
	iframe = new JInternalFrame("Text Display with Markables", true, false, true, true);
	iframe.setSize(new Dimension(500, 570));
 	addTextAreaContent(iframe);
	/* Find the element for initial highlight */
	if (markable_list.size()!=0) {
	    changed = (NOMElement) markable_list.getFirst();
	}
	desktop.add(iframe);

	noteIFrame = new JInternalFrame("Note", true, false, true, true);
	noteIFrame.setSize(new Dimension(300, 100));
	noteIFrame.setLocation(new Point(500, 625));
	makeNoteTypeIn(noteIFrame);
	desktop.add(noteIFrame);


	desktop.setSize(new Dimension(800, 710));
        desktop.setActionMap(map.getActionMap());
        desktop.setInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT,map.getInputMap());
	map.addKeyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, 0), new Skip(false));
	map.addKeyStroke(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, 0), new Skip(true));
	//String key = map.addKeyStroke("" + KeyEvent.VK_LEFT, new Skip(true));

	frame.getContentPane().add(desktop);
	frame.setSize(new Dimension(810, 780));
	frame.setVisible(true);

	if (changed!=null) {
	    nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(changed));
	    showStatus(changed);
	    showNote(changed);
	}

	List signals = controlData.getSignals();
	if (signals!=null && signals.size()>0) {
	    NSignal sig = (NSignal) signals.get(0); // first signal;
	    if (sig.getMediaType()==NSignal.AUDIO_SIGNAL) {
		String audio_filename=controlData.getSignalPath() + File.separator +
		    observationname + "." + sig.getName() + "." + sig.getExtension();
		System.out.println("Audio file: " + audio_filename);
		NITEAudioPlayer audio = new NITEAudioPlayer(new File(audio_filename), (DefaultClock) niteclock);
		audio.setLocation(new Point(500, 370));
		desktop.add(audio);
	    }
	}
    }

    /* Display an ontology tree as an NTree */
    private NTree showOntologyTree(String element, String attribute) {
	List elist = null;
	try {
	    elist = searchEngine.search(nom, "($a " + element + ")");
	} catch (Throwable e) {
	    e.printStackTrace();
	    System.exit(0);
	}
	if (elist == null || elist.size() <= 1) {
	    System.err.println("NO '" + element + "'s IN CORPUS!");
	    System.exit(0);
	}


	NOMElement etop = (NOMElement) ((ArrayList)elist.get(1)).get(0);
	NTreeNode root = new NTreeNode(etop.getAttribute(attribute).getStringValue());
	linktyperootpath=new TreePath(root);
	NOMObjectModelElement nome= new NOMObjectModelElement(etop);
	root.setDataElement((ObjectModelElement)nome);
	showTree(etop, root, attribute);

	lttree = new NTree(root);
	// Open all the root's children
	Enumeration kiden=root.children();
	while (kiden.hasMoreElements() ) {
	    DefaultMutableTreeNode jtn=(DefaultMutableTreeNode)kiden.nextElement();
	    TreeNode ktn=null;
	    if (jtn.getChildCount()!=0) { ktn=jtn.getFirstChild(); }
	    if (ktn==null) { continue; }
	    TreePath path = new TreePath(((DefaultTreeModel) lttree.getModel()).getPathToRoot(ktn));
	    if (!lttree.isVisible(path)) {
		lttree.scrollPathToVisible(path);
	    }
	}
	return lttree;
    }

    private void showNote(NOMElement nomel) {
	String text = (String)nomel.getComment();
	if (text==null) { text=""; }
	noteTextField.setText(text);
    }	

    private void showTree(NOMElement source_element, NTreeNode tree_node, 
			  String attribute) {
	if (source_element==null) { return; }
	NTreeNode newnode=null;
	if (source_element.getParents()!=null) {
	    newnode = new NTreeNode(source_element.getAttribute(attribute).getStringValue());
	    NOMObjectModelElement nome= new NOMObjectModelElement(source_element);
	    newnode.setDataElement((ObjectModelElement)nome);
	    tree_node.addNode(newnode);
	} else {
	    newnode=tree_node;
	}
	if (source_element.getChildren() == null) { return; }
	for (Iterator kit=source_element.getChildren().iterator(); kit.hasNext(); ) { 
	    NOMElement kid = (NOMElement) kit.next();
	    showTree(kid, newnode, attribute);
	}
    }

    /** put up a notes section; when the user hits return, it adds
     * the note for the currently selected markable.
     */
    private void makeNoteTypeIn(JInternalFrame f) {
	JPanel panel = new JPanel();
	JButton nbut = new JButton(NOTEBUTTON);
	//nbut.setFont(new Font("Serif", Font.PLAIN, textsize));
	nbut.addActionListener(this);
	panel.add(nbut);
	f.getContentPane().add(panel);
	noteTextField = new JTextField(20);
        noteTextField.addActionListener(this);
        panel.add(noteTextField);
	f.setVisible(true);
    }


    /* Add the buttons to set markable status */
    private void addButtons(JPanel panel) {
	NElement nell = controlData.getElementByName(MARKABLE);
	NAttribute nat = nell.getAttributeByName(STATUS);
	buttonmap = new HashMap();
	String shortcuts="1234567890QWERTYUIOP";
	
	List vals = nat.getEnumeratedValues();
	Iterator nit = vals.iterator();
	int i=0;
	while (nit.hasNext()) {
	    String val = (String) nit.next();
	    JButton nbut = new JButton(val);
	    //Action act = new SetValue(val);
	    //nbut.setAction(act);
	    //map.addKeyStroke(shortcuts.substring(i,i+1), act);
	    i++;
	    nbut.addActionListener(this);
	    buttonmap.put(val, nbut);
	    panel.add(nbut);
	}
    }

    /* Add the buttons to set markable status type *
    private void addStatusTypeButtons(JPanel panel) {
	NElement nell = controlData.getElementByName(MARKABLE);
	NAttribute nat = nell.getAttributeByName(STATUSTYPE);
	sbuttonmap = new HashMap();
	List vals = nat.getEnumeratedValues();
	Iterator nit = vals.iterator();
	JButton nbut=null;
	while (nit.hasNext()) {
	    String val = (String) nit.next();
	    nbut = new JButton(val);
	    sbuttonmap.put(val, nbut);
	    nbut.addActionListener(this);
	    panel.add(nbut);
	}
	newbut=nbut;
    }
    */

    /* Add buttons for link actions */
    private void addLinkButtons(JPanel panel) {
	JButton nbut = new JButton(newlink);
	nbut.addActionListener(this);
	nbut.setToolTipText(newlinktext);
	panel.add(nbut);
	nbut = new JButton(deletelink);
	nbut.addActionListener(this);
	panel.add(nbut);
	nbut = new JButton(antecedent);
	nbut.addActionListener(this);
	panel.add(nbut);
	//	nbut = new JButton(referent);
	//	nbut.addActionListener(this);
	//	panel.add(nbut);
	//nbut = new JButton(linktype);
	//nbut.addActionListener(this);
	//panel.add(nbut);
    }

    /* Clear a display element frame ready for redisplay */
    private void clearFrame(JInternalFrame frame) {
	frame.setVisible(false);
	frame.getContentPane().removeAll();
    }

    /** put buttons corresponding to skip forwards and skip backwards
     */
    private void makeSkipAndBackButtons(JInternalFrame f) {
	JPanel skipbuttonpanel = new JPanel();
	addSkipButton(skipbuttonpanel);
	addBackButton(skipbuttonpanel);
	f.getContentPane().add(skipbuttonpanel);
	f.setVisible(true);
    }

    /** Add an extra button on another panel in the same frame
     * as the markable coding buttons that skips forward, leaving
     * the coding exactly as it was.  (The coding interface moves on
     * to the next markable when one is coded).
     */
    private void addSkipButton(JPanel panel) {
	JButton nbut = new JButton(FORWARD);
	//nbut.setFont(new Font("Serif", Font.PLAIN, textsize));
	nbut.addActionListener(this);
	panel.add(nbut);
    }
    
    /** Add an extra button on another panel in the same frame
     * as the markable coding buttons that skips backward, leaving
     * the coding exactly as it was.  
     */
    private void addBackButton(JPanel panel) {
	JButton nbut = new JButton(BACKWARD);
	//nbut.setFont(new Font("Serif", Font.PLAIN, textsize));
	nbut.addActionListener(this);
	panel.add(nbut);
    }

    /* make some styles using a TextStyleHandler */
    private void makeStyles() {
	for (int i=0; i<stylenamelist.length; i++) {
	    TextStyleHandler stylehandler = new TextStyleHandler();
	    stylehandler.init("", null);
	    stylehandler.setName(stylenamelist[i]);
	    stylehandler.makeNewStyle();
	    stylehandler.setStyle( stylehandler.getColours(stylehandler.getStyle(), stylenamelist[i], stylenamelist[i]));
	    StyleConstants.setFontSize(stylehandler.getStyle(), 16);
	    StyleConstants.setBold(stylehandler.getStyle(), true);
	    stylelist[i]=stylehandler.getStyle();
	} 
    }
    
    /* display or redisplay the main text area */
    private void addTextAreaContent(JInternalFrame frame ) {
	if (nta!=null && search!=null) { search.deregisterResultHandler(nta); }
	if (nta!=null && niteclock!=null) { niteclock.deregisterTimeHandler(nta); }
	nta = new NTextArea();
	if (search != null) { search.registerResultHandler(nta); }
	if (niteclock!=null) { niteclock.registerTimeHandler(nta); }
	high=null;
	//	ToolTipManager.sharedInstance().registerComponent(nta);
	nta.addMouseListener(this);
	for (int i=0; i<stylenamelist.length; i++) {
	    nta.addStyle(stylenamelist[i], stylelist[i]);
	}
	List elist = null;
	
	try {
	    //elist = searchEngine.search(nom, "($a parse)");
	    elist = searchEngine.search(nom, "($a turn)");
	} catch (Throwable e) {
	    e.printStackTrace();
	    System.exit(0);
	}
	if (elist == null || elist.size() <= 1) {
	    //System.err.println("NO parse ELEMENTS IN CORPUS!");
	    System.err.println("NO turn ELEMENTS IN CORPUS!");
	    System.exit(0);
	}

        /* remove first (dud) result */
	elist.remove(0);
        /* sort in time order */
	Collections.sort(elist, mycomp);

	/* The first thing on the list returned by the search engine
	 * is an entry containing the names of the variables */
	for (int i=0; i<elist.size(); i++) {
	    List l = (List) elist.get(i);
	    NOMElement nwe = (NOMElement) l.get(0);
	    NOMObjectModelElement nome= new NOMObjectModelElement(nwe);
	    NTextElement nte = new NTextElement(nwe.getAgentName() + ": ", "red");
	    nte.setDataElement((ObjectModelElement)nome);
	    nta.addElement(nte);
	    showWords(nwe);
	    nte = new NTextElement("\n\n", null);    
	    nta.addElement(nte);
	}

	JScrollPane pane = new JScrollPane(nta);
	frame.getContentPane().add(pane);
	frame.setVisible(true);
    }

    //new JAN 08
   private String getTextForTranscriptionElement(NOMElement nme) {
	String name = nme.getName();
	String content = "";
	
	if (name.equalsIgnoreCase(PUNC)) {
        	content = nme.getText();
		if (content==null) { return ""; }
		return content.trim();
	} else if (name.equalsIgnoreCase(SIL)) {
	    return "";
        } else if (name.equalsIgnoreCase(WORD)) {
        	NOMAttribute orth = nme.getAttribute(ORTH);
		if (orth==null) { return ""; }
		content = orth.getStringValue();
		if (content==null) { return ""; }
		return content.trim();
	} else {
	    return "[" + name + "]";
	}
	
    }


    /* show the words in the corpus, recursively descending through
       the syntax layers and checking for markables at each stage. */
    private void showWords(NOMElement nwe) {
	if (nwe==null || nwe.getChildren()==null) { return; }
	showMarkableParentsStart(nwe);
	for (Iterator kit=nwe.getChildren().iterator(); kit.hasNext(); ) {
	    NOMElement kid=(NOMElement)kit.next();
	    if (kid.getName().equalsIgnoreCase("word")) {
		showMarkableParentsStart(kid);
		NOMObjectModelElement nome= new NOMObjectModelElement(kid);
		if (!markable_stack.empty()) {
		    NOMObjectModelElement parnome = (NOMObjectModelElement)markable_stack.peek();
		    //JAN 08
		    //NTextElement nte=new NTextElement(kid.getText() + " ", null, kid.getStartTime(), kid.getEndTime(), nome);
		    NTextElement nte=new NTextElement(getTextForTranscriptionElement(kid) + " ", null, kid.getStartTime(), kid.getEndTime(), nome);
		    nte.setEditElement(parnome);
		    nta.addElement(nte);
		} else {
		 //JAN 08
		    //nta.addElement(new NTextElement(kid.getText() + " ", null, kid.getStartTime(), kid.getEndTime(), nome));
		    nta.addElement(new NTextElement(getTextForTranscriptionElement(kid) + " ", null, kid.getStartTime(), kid.getEndTime(), nome));
		}
		showMarkableParentsEnd(kid);
	    } else if (kid.getName().equalsIgnoreCase("trace")) {
		showMarkableParentsStart(kid);
		NOMObjectModelElement nome= new NOMObjectModelElement(kid);
		if (!markable_stack.empty()) {
		    NOMObjectModelElement parnome = (NOMObjectModelElement)markable_stack.peek();
		    nta.addElement(new NTextElement("<trace> ", null, nome, parnome));
		} else {
		    nta.addElement(new NTextElement("<trace> ", null, nome));
		}
		showMarkableParentsEnd(kid);
	    } else {
		showWords(kid);
	    }
	}
	showMarkableParentsEnd(nwe);
    }

    /* display any open-markable elements in the appropriate colour */
    private int countMarkableKids(NOMElement nwe) {
	int markids=0;
	if (nwe==null || nwe.getChildren()==null) { return markids; }
	for (Iterator kit=nwe.getChildren().iterator(); kit.hasNext(); ) {
	    NOMElement kid = (NOMElement)kit.next();
	    if (kid.getPointersTo()==null) continue;
	    for (Iterator pit=nwe.getPointersTo().iterator(); pit.hasNext(); ) {
		NOMPointer point=(NOMPointer)pit.next();
		if (point==null) continue;
		NOMElement parent = point.getFromElement();
		if (parent.getName().equalsIgnoreCase(MARKABLE)) {
		    markids+=1;
		}
	    }
	}
	return markids;
    }


    /* display any open-markable elements in the appropriate colour */
    private void showMarkableParentsStart(NOMElement nwe) {
	NOMElement parent=nwe.findAncestorNamed(MARKABLE);
	if (parent==null) { return; }
	NOMElement presib = (NOMElement)parent.getChildren().get(0);
	if (presib!=nwe) { return; }
	marknest++;
	String status = "";
	if (parent.getAttribute(STATUS)!=null) {
	    status=parent.getAttribute(STATUS).getStringValue() + " ";
	}
	int index=marknest;
	if (marknest >= stylenamelist.length) { 
	    index=(marknest % stylenamelist.length)+1;
	    if (index == stylenamelist.length) { index=1; }
	}
	String sname = stylenamelist[index];
	NTextElement nte = new NTextElement(statusStartString(status), sname);
	NOMObjectModelElement nome= new NOMObjectModelElement(parent);
	nome.setDisplayedAttribute(STATUS);
	nte.setDataElement((ObjectModelElement)nome);
	nta.addElement(nte);
	markable_stack.push(nome);
	markable_list.add(parent);
	if (parent==changed) { high=nte; }
    }

    /* display any close-markable elements in the appropriate colour */
    private void showMarkableParentsEnd(NOMElement nwe) {
	NOMElement parent=nwe.findAncestorNamed(MARKABLE);
	if (parent==null) { return; }
	List kds = parent.getChildren();
	NOMElement postsib = (NOMElement)kds.get(kds.size()-1);
	if (postsib!=nwe) { return; }
	int index=marknest;
	if (marknest >= stylenamelist.length) { 
	    index=(marknest % stylenamelist.length)+1;
	    if (index == stylenamelist.length) { index=1; }
	}
	String sname = stylenamelist[index];
	String status = "";
	if (parent.getAttribute(STATUS)!=null) {
	    status=parent.getAttribute(STATUS).getStringValue() + " ";
	}
	NTextElement nte = new NTextElement(statusEndString(status), sname);
	NOMObjectModelElement nome= new NOMObjectModelElement(parent);
	nte.setDataElement((ObjectModelElement)nome);
	nta.addElement(nte);
	nte = new NTextElement(" ", null);
	nta.addElement(nte);
	markable_stack.pop();
	marknest--;
    }
    
    /* this is where we could select special chars to begin certain
       types of markable */
    private String statusStartString(String status) {
	String retstr="(";
	if (status==null || status.equals("")) { retstr += UNCODED_STRING; }
	/*
	if (status!=null) {
	    if (status.length() > 1) { 
		if (status.lastIndexOf("-")==0) {
		    status = status.substring(0,2);
		} else {
		    status = status.substring(0,1);
		}
	    }
	    retstr += status;
	}
	*/
	return retstr + " ";
    }

    /* this is where we could select special chars to end certain
       types of markable */
    private String statusEndString(String status) {
	String retstr=")";
	return retstr;
    }

    /* display the list of links */
    private void addLinkPaneContent(JInternalFrame frame ) {
	List elist = null;
	linkmap=new HashMap();
	try {
	    elist = searchEngine.search(nom, "($a " + TRIGGER + ")");
	} catch (Throwable e) {
	    e.printStackTrace();
	    System.exit(0);
	}
	if (elist == null || elist.size() <= 1) {
	    System.err.println("NO LINK ELEMENTS IN CORPUS!");
	    //	    System.exit(0);
	}

	NTreeNode root = new NTreeNode("Links");

	for (int i=1; i<elist.size(); i++) {
	    List l = (List) elist.get(i);
	    showLink(root, (NOMElement)l.get(0));
	}

	ntree = new NTree(root);
	ntree.addTreeSelectionListener(this);
	ntree.setCellRenderer(new NTreeCellRenderer());
	rootpath=new TreePath(root);
	JScrollPane pane = new JScrollPane(ntree);
	frame.getContentPane().add(pane);
	frame.setVisible(true);
	if (changed_treenode!=null) {
	    // select the previous selection
	    TreePath path = new TreePath(((DefaultTreeModel) ntree.getModel()).getPathToRoot(changed_treenode));
	    ntree.addSelectionPath(path);
	}
	//	ntree.visitAllNodesForData();
    }

    /* Display a link by adding it to the NTree */
    private void showLink(NTreeNode parent, NOMElement link) {
	String text = link.getID() + " ";
	NOMPointer np=link.getPointerWithRole(linktypeatt);
	if (np!=null) {
	    text += "(" + np.getToElement().getAttribute(linktypeatt).getStringValue() + ")"; 
	} else {
	    text += "(uncoded)";
	}
	NTreeNode newnode = new NTreeNode(text);
	NOMObjectModelElement nome= new NOMObjectModelElement(link);
	newnode.setDataElement((ObjectModelElement)nome);
	if (link.getPointers()!=null) {
	    for (Iterator pit=link.getPointers().iterator(); pit.hasNext(); ) {
		NOMPointer pointer=(NOMPointer)pit.next();
		if (pointer.getRole().equalsIgnoreCase(linktypeatt)) { continue; }
		NTreeNode nnode = new NTreeNode(pointer.getRole());
		if (pointer.getRole().equalsIgnoreCase(refstr)) { nnode.textcolour=refcolour; }
		if (pointer.getRole().equalsIgnoreCase(antestr)) { nnode.textcolour=antecolour; }
		nome = new NOMObjectModelElement(pointer.getToElement());
		Set linklist = (Set) linkmap.get(pointer.getToElement());
		if (linklist==null) { linklist = (Set) new HashSet(); }
		linklist.add(nome);
		linkmap.put(pointer.getToElement(),linklist);
		nnode.setDataElement((ObjectModelElement)nome);
		newnode.addNode(nnode);
	    }
	}
	parent.addNode(newnode);
	if (link==changed_link) {
	    changed_treenode=newnode;
	}
    }
    
    /** Start the application. */
    public static void main(String[] args) {
	String corpus=null;
	String observation=null;

	if (args.length < 2 || args.length > 4) { usage(); }
	for (int i=0; i<args.length; i++) {
	    String flag=args[i];
	    if (flag.equals("-corpus") || flag.equals("-c")) {
		i++; if (i>=args.length) { usage(); }
		corpus=args[i];
	    } else if (flag.equals("-observation") || flag.equals("-o")) {
		i++; if (i>=args.length) { usage(); }
		observation=args[i];
	    } else {
		usage();
	    }
	}
	if (corpus == null) { usage(); }
	
	SwitchboardKontrast m = new SwitchboardKontrast(corpus, observation);
    }
    
    private static void usage() {
	System.err.println("Usage: java SwitchboardKontrast -corpus metadata-filename [ -observation observation-name ]");
	System.exit(0);
    }

    /** This handles all the button-presses etc for the interface */
    public void actionPerformed(ActionEvent ae) {
	//	System.out.println("Action: " + ae.getActionCommand());
	String act = ae.getActionCommand();
	Object src = ae.getSource();
	clearIcons();
	clearStatusTypeIcons();
	if (act.equalsIgnoreCase(linktype)) {
	    changeLinkType();
	    return;
	}
	List sel = nta.getHighlightedModelElements(NTextArea.SELECTION_HIGHLIGHTS);
	if (sel == null || sel.size()==0) {
	    System.out.println("Nothing selected in text area");
	} else {
	    //	    System.out.println("Selected: " + sel);
	    Iterator sit = sel.iterator();
	    while (sit.hasNext()) {
		NOMObjectModelElement el = (NOMObjectModelElement) sit.next();
		if (el==null) { continue; }
		if (!el.getName().equalsIgnoreCase(MARKABLE)) { 
		    System.out.println("Please select a markable");
		    return; 
		}
		changed=el.getElement();
		//System.out.println("Selected: " + (changed).getID() + "; ACTION: " + act + "; SOURCE: " + src);
		if (act.equalsIgnoreCase(newlink)) {
		    addLinkWithReferent(changed, (ObjectModelElement)el);
		    return;
		} else if (act.equalsIgnoreCase(deletelink)) {
		    deleteLink();
		    return;
		} else if (act.equalsIgnoreCase(antecedent)) {
		    changeLinkAntecedent(changed);
		    return;
		} else if (act.equalsIgnoreCase(referent)) {
		    changeLinkReferent(changed);
		    return;
		} else if (act.equals(BACKWARD)) {
		    skip(false);
		    return;
		} else if (act.equals(FORWARD)) {
		    skip(true);
		    return;
		} else if (act.equals(NOTEBUTTON) || src==noteTextField) {
		    String note = noteTextField.getText();
		    changed.setComment(note);
		    return;
		}

		String status=null;
		if (changed.getAttribute(STATUS)!=null) {
		    status=changed.getAttribute(STATUS).getStringValue() + " ";
		}
		String oldstring=statusStartString(status);
		try {
		    changed.setStringAttribute(STATUS, act);
		} catch (NOMException nex) {
		    System.err.println("FAILED TO SET ATTRIBUTE VALUE!\n");
		    return;
		}
		String newstring=statusStartString(act);
		if (!oldstring.equals(newstring)) {
		    nta.redisplayElement(el, oldstring, newstring);
		}
		nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(changed));
	    }
	    skip(true);
	}
	showStatus(changed);
	showNote(changed);
    }


    /* add a new link to the corpus with the referent being the
       currently highlighted markable */
    private void addLinkWithReferent(NOMElement referent, ObjectModelElement el) {
	try {
	    NOMWriteElement nwe = new NOMWriteAnnotation(nom, TRIGGER, observationname, (String)null);
	    nwe.addPointer(new NOMWritePointer(nom, refstr, nwe, referent));
	    nwe.addToCorpus();
	    //	    ntree.redisplayElement((ObjectModelElement)el);
	    changed_link=nwe;
	    clearFrame(iframe3);
	    addLinkPaneContent(iframe3);
	    WAITING_FOR_ANTECEDENT_MODE=true;
	} catch (NOMException nex) {
	    nex.printStackTrace();
	}
    }

    /* delete a link from the corpus */
    private void deleteLink() {
	NOMWriteElement link = findLink();
	if (link==null) { return; }
	if (!link.getName().equals(TRIGGER)) { return; }
	try {
	    link.getParentInFile().deleteChild(link);
	    changed_link=null;
	    clearFrame(iframe3);
	    addLinkPaneContent(iframe3);
	} catch (NOMException nex) {
	    nex.printStackTrace();
	}
    }

    /* change the referent of the link (by changing the pointer with
       role "referent") and redisplay the links */
    private void changeLinkReferent(NOMElement referent) {
	NOMWriteElement link = findLink();
	if (link ==null) { return; }
	NOMPointer ref = link.getPointerWithRole(refstr);
	if (ref == null) {
		try {
	      link.addPointer(new NOMWritePointer(nom, refstr, link, referent));
		} catch (NOMException nex) {
			  nex.printStackTrace();
		}
	    } else {
	    	try {
	    ref.setToElement(referent);
	    	} catch (NOMException nex) {
			  nex.printStackTrace();
		}
	}
	changed_link=link;
	clearFrame(iframe3);
	addLinkPaneContent(iframe3);
    }

    /* change the antecedent of the link (by changing the pointer with
       role "antecedent") and redisplay the links */
    private void changeLinkAntecedent(NOMElement antecedent) {
	NOMWriteElement link = findLink();
	if (link==null) { return; }
	NOMPointer ref = link.getPointerWithRole(antestr);
	if (ref == null) {
		try {
	    link.addPointer(new NOMWritePointer(nom, antestr, link, antecedent));
	    		} catch (NOMException nex) {
			  nex.printStackTrace();
		}
	} else {
		try {
	    ref.setToElement(antecedent);
	    	    	} catch (NOMException nex) {
			  nex.printStackTrace();
		}
	}
	changed_link=link;
	clearFrame(iframe3);
	addLinkPaneContent(iframe3);
    }

    /* change the type of the link (by changing the pointer into the
       ontology) and redisplay the links */
    private void changeLinkType() {
	NOMWriteElement link = findLink();
	NOMWriteElement linktype = findLinkType();
	if (link ==null || linktype==null || linktype.getAttribute(linktypeatt)==null) { 
	    return;
	}
	NOMPointer ref = link.getPointerWithRole(linktypeatt);
	if (ref == null) {
		try {
	    link.addPointer(new NOMWritePointer(nom, linktypeatt, link, linktype));
	    	    	} catch (NOMException nex) {
			  nex.printStackTrace();
		}
	} else {
		try {
	    ref.setToElement(linktype);
	    	    	} catch (NOMException nex) {
			  nex.printStackTrace();
		}
	}
	changed_link=(NOMElement)link;
	clearFrame(iframe3);
	addLinkPaneContent(iframe3);
    }

    /** find the highlighted link */
    private NOMWriteElement findLink() {
	Set sel = ntree.getSelectedTreeNodes();
	if (sel == null || sel.iterator()==null || !sel.iterator().hasNext()) {
	    System.out.println("Please select a link!");
	    return null;
	} else {
	    NTreeNode ntn = (NTreeNode) sel.iterator().next();
	    NOMObjectModelElement el = (NOMObjectModelElement) ntn.getDataElement();
	    if (el==null) { return null; }
	    return (NOMWriteElement)el.getElement();
	}
    }

    /** find the highlighted link type*/
    private NOMWriteElement findLinkType() {
	Set sel = lttree.getSelectedTreeNodes();
	if (sel == null || sel.iterator()==null || !sel.iterator().hasNext()) {
	    System.out.println("Please select a link type!");
	    return null;
	} else {
	    NTreeNode ntn = (NTreeNode) sel.iterator().next();
	    NOMObjectModelElement el = (NOMObjectModelElement) ntn.getDataElement();
	    if (el==null) { return null; }
	    return (NOMWriteElement)el.getElement();
	}
    }

    /** Save the NOM using the parameters set on the interface
     */
    public class SaveAction extends AbstractAction {
	public SaveAction() {
	    super("Save corpus");
	}
	public void actionPerformed(ActionEvent ev) {
	    try {
		// controlData.writeMetaData(corpusname);
		nom.serializeCorpusChanged();
	    } catch (Exception e) {
		e.printStackTrace();
	    }
	}
    }

    /** Save the changes to the corpus using the dialogue provided
     */
    public class SaveAsAction extends AbstractAction {
	public SaveAsAction() {
	    super("Save as...");
	}
	
	public void actionPerformed(ActionEvent ev) {
	    CorpusSaveOptions cso = new CorpusSaveOptions(nom);
	}
    }

    /** Change observation but check for changes first */
    public class OpenObservationAction extends AbstractAction {
	public OpenObservationAction() {
	    super("Open...");
	}
	public void actionPerformed(ActionEvent ev) {
	    CheckSave co = new CheckSave(nom); 
	    if (!(co.popupDialog()==JOptionPane.CANCEL_OPTION)) {
		ChooseObservation cob = new ChooseObservation(controlData);
		observationname=cob.popupDialog();
		System.out.println("Opening " + observationname);
		nom.clearData();
		markable_list = new LinkedList();
		NObservation obs = controlData.getObservationWithName(observationname);
		try {
		    nom.loadData(obs);
		} catch (NOMException nex) {
		    nex.printStackTrace();
		}
		clearFrame(iframe);
		addTextAreaContent(iframe);
		nta.clearHighlights(NTextArea.SELECTION_HIGHLIGHTS);
		changed = (NOMElement)markable_list.getFirst();
		nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(changed));
		//		nta.setHighlighted(new NOMObjectModelElement(changed), MetalLookAndFeel.getTextHighlightColor());
		clearIcons();
	    }
	}
    }

    /** Save the NOM using the parameters set on the interface
     */
    public class RedisplayAction extends AbstractAction {
	public RedisplayAction() {
	    super("Redisplay...");
	}

	public void actionPerformed(ActionEvent ev) {
	    clearIcons();
	    clearStatusTypeIcons();
	    clearFrame(iframe);
	    addTextAreaContent(iframe);
	    nta.clearHighlights(NTextArea.SELECTION_HIGHLIGHTS);
	    if (changed!=null) {
		nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(changed));
		//nta.setHighlighted(new NOMObjectModelElement(changed), MetalLookAndFeel.getTextHighlightColor());
		showStatus(changed);
		showNote(changed);
	    }
	}
    }

    /** Print text area */
    public class PrintAction extends AbstractAction {
	public PrintAction() {
	    super("Print...");
	}
	public void actionPerformed(ActionEvent ev) {
	    nta.printTextArea();
	}
    }
    
    /** A very simple "exit" action - just checks for changes and saves if so */
    public class ExitAction extends AbstractAction {
	public ExitAction() {
	    super("Exit");
	}
	public void actionPerformed(ActionEvent ev) {
	    CheckSave co = new CheckSave(nom); 
	    if (!(co.popupDialog()==JOptionPane.CANCEL_OPTION)) {
		System.exit(0);
	    }
	}
    }

    /** A "search" action - pops up a search GUI */
    public class SearchAction extends AbstractAction {
	public SearchAction() {
	    super("Search...");
	}
	public void actionPerformed(ActionEvent ev) {
	    search.popupSearchWindow();
	}
    }


    /** TreeSelectionEvents - highlight the referent and antecedent on
        the text area */
    public void valueChanged (TreeSelectionEvent tse) {
	/*
	if (linktyperootpath.isDescendant(tse.getNewLeadSelectionPath())) {
	    linkTypeSelected(tse);
	    return;
	}
	*/
	NOMElement nel = findLink();
	WAITING_FOR_ANTECEDENT_MODE=false;
	if (nel==null) { return; }
	nta.clearHighlights(NTextArea.USER_HIGHLIGHTS);
	clearIcons();
	clearStatusTypeIcons();
	if (nel.getName().equalsIgnoreCase(TRIGGER)) {
	    for (Iterator pit=nel.getPointers().iterator(); pit.hasNext(); ) {
		NOMPointer pointer=(NOMPointer)pit.next();
		if (pointer.getRole().equalsIgnoreCase(antestr)) {
		    highlightAntecedent(pointer.getToElement());
		} else if (pointer.getRole().equalsIgnoreCase(refstr)) {
		    highlightReferent(pointer.getToElement());
		}
	    }
	    changed_link=nel;
	} else {
	    // this is a bit cheesy - testing the string because we
	    // can't associate the Pointer with the text element
	    String path = ((TreeNode)tse.getNewLeadSelectionPath().getLastPathComponent()).toString();
	    if (path.indexOf(refstr) >= 0) {
		highlightReferent(nel);
	    } else if (path.indexOf(antestr) >= 0) {
		highlightAntecedent(nel);
	    } else {
		nta.setHighlighted(NTextArea.USER_HIGHLIGHTS, new NOMObjectModelElement(nel));
		//		nta.setHighlighted((ObjectModelElement)new NOMObjectModelElement(nel),
		//			   MetalLookAndFeel.getTextHighlightColor());
	    }
	}
    }

    private void highlightAntecedent(NOMElement ante) {
	nta.setHighlighted(NTextArea.USER_HIGHLIGHTS, new NOMObjectModelElement(ante));
	//	nta.setHighlighted((ObjectModelElement)new NOMObjectModelElement(ante),
	//			   antecolour);
    }

    private void highlightReferent(NOMElement ref) {
	nta.setHighlighted(NTextArea.USER_HIGHLIGHTS, new NOMObjectModelElement(ref));
	//	nta.setHighlighted((ObjectModelElement)new NOMObjectModelElement(ref),
	// refcolour);
    }

    private void linkTypeSelected(TreeSelectionEvent tse) {
	//	System.out.println("Link type selected");
	changeLinkType();
    }

    private void clearIcons() {
	for (Iterator bit=buttonmap.values().iterator(); bit.hasNext(); ) {
	    JButton but = (JButton)bit.next();
	    if (but.getIcon() != null) {
		but.setIcon(null);
		but.repaint();
	    }
	}
    }

    private void clearStatusTypeIcons() {
	for (Iterator bit=sbuttonmap.values().iterator(); bit.hasNext(); ) {
	    JButton but = (JButton)bit.next();
	    if (but.getIcon() != null) {
		but.setIcon(null);
		but.repaint();
	    }
	}
    }

    private void clearLinks() {
	ntree.clearHighlights();
    }


    private void showStatus(NOMElement nel) {
	String st = (String)nel.getAttributeComparableValue(STATUS);
	if (st==null) { return; }
	JButton mybutton=null;
	if (buttonmap.get(st) != null) { mybutton = (JButton)buttonmap.get(st);	}
	if (mybutton!=null) {
	    mybutton.setIcon(highlight_image);
	    mybutton.repaint();
	}
    }

    private void showStatusType(NOMElement nel) {
	String st = (String)nel.getAttributeComparableValue(STATUSTYPE);
	if (st==null) { return; }
	JButton mybutton=null;
	if (sbuttonmap.get(st) != null) { mybutton = (JButton)sbuttonmap.get(st);	}
	if (mybutton!=null) {
	    mybutton.setIcon(highlight_green_image);
	    mybutton.repaint();
	}
    }

    private void showLinks(NOMElement nel) {
	if (linkmap.get(nel)==null) { return; }
	Set links = (Set)linkmap.get(nel);
	//	System.out.println("Iterate: " + nel.getID());
	for (Iterator lit=links.iterator(); lit.hasNext(); ) {
	    ObjectModelElement lel = (ObjectModelElement)lit.next();
	    ntree.setHighlighted(lel, antecolour);
	}
    }

    /* Mouse listener implementation */
    public void mouseClicked(MouseEvent e) {
	//	System.out.println("Click");
    }
    public void mouseEntered(MouseEvent e) {

    }
    public void mouseExited(MouseEvent e) {

    }
    public void mousePressed(MouseEvent e) {
	//	System.out.println("Press");
    }
    public void mouseReleased(MouseEvent e) {
	//	System.out.println("Release");
	List sel = nta.getHighlightedModelElements(NTextArea.SELECTION_HIGHLIGHTS);
	if (sel == null || sel.size()==0) {
	    System.out.println("Nothing selected in text area");
	    return;
	} 
	clearIcons();
	clearStatusTypeIcons();
	clearLinks();
	for (Iterator sit = sel.iterator(); sit.hasNext(); ) {
	    NOMObjectModelElement el = (NOMObjectModelElement) sit.next();
	    if (el.getName().equals(MARKABLE)) {
		//System.out.println("Selection: " + el.getElement().getID());
		NOMElement nel= el.getElement();
		if (WAITING_FOR_ANTECEDENT_MODE) {
		    changeLinkAntecedent(nel);		    
		    WAITING_FOR_ANTECEDENT_MODE=false;
		}
		showStatus(nel);
		showNote(nel);
		showLinks(nel);
	    }
	}
    }

    private ImageIcon loadIcon(String name) {
	try {
	    // get resource
	    URL resource = ClassLoader.getSystemClassLoader().getResource(name);
	    if( resource == null ){ resource = getClass().getResource(name); }
	    if( resource == null ){ 
		System.err.println("WARNING: Failed to load image " + name);
		return null;
	    }
	    // return image
	    return new ImageIcon( frame.getToolkit().getImage(resource) );
	} catch(Exception e) { 
	    System.err.println("WARNING: Failed to load image " + name);
	    return null;
	}
    }

    /* NOMView implementation */

  /*  public int getViewLevel() {
	return NOMView.DETAILED_VIEW;
    }

    public void handleChange() {
	System.out.println("NOM: Arbitrary change");
    }

    public void handleChange(NOMEdit edit) {
	System.out.println("NOM: " + edit.toString());
    }
    */

    private void skip(boolean forward) {
	int ind = markable_list.indexOf(changed);
	nta.clearHighlights(NTextArea.SELECTION_HIGHLIGHTS);
	clearIcons();
	if (forward) {
	    if (ind != -1 && ind != markable_list.size() - 1) {
		changed = (NOMElement) markable_list.get(ind + 1);
	    }
	} else {
	    if (ind != -1 && ind != 0) {
		changed = (NOMElement) markable_list.get(ind - 1);
	    }
	}
	nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(changed));
	showStatus(changed);
	showNote(changed);
    }

    private class Skip extends AbstractAction {
        private boolean forward=true;

        public Skip(boolean forwards) {
	    this.forward = forwards;
        }

        public void actionPerformed(ActionEvent e) {
	    skip(forward);
        }
    }

    private class SetValue extends AbstractAction {
        private String value=null;

        public SetValue(String val) {
	    this.value = val;
	    System.out.println("NEW button: value is: " + val);
        }

        public void actionPerformed(ActionEvent e) {
	    List sel = nta.getHighlightedModelElements(NTextArea.SELECTION_HIGHLIGHTS);
	    if (sel == null || sel.size()==0) {
		System.out.println("Nothing selected in text area");
		return;
	    }
	    NOMObjectModelElement el = (NOMObjectModelElement) sel.get(0);
	    String status=null;
	    if (changed.getAttribute(STATUS)!=null) {
		status=changed.getAttribute(STATUS).getStringValue() + " ";
	    }
	    String oldstring=statusStartString(status);
	    try {
		changed.setStringAttribute(STATUS, value);
	    } catch (NOMException nex) {
		System.err.println("FAILED TO SET ATTRIBUTE VALUE!\n");
		return;
	    }
	    String newstring=statusStartString(value);
	    if (!oldstring.equals(newstring)) {
		nta.redisplayElement(el, oldstring, newstring);
	    }
        }
    }

}
